/*
    *
    * This file is a part of Libcsys.
    * Library for managing drive and getting system resource information in real time for CuboCore Application Suite
	* Copyright 2019 CuboCore Group
    *

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, vsit http://www.gnu.org/licenses/.
    *
*/

#include <unistd.h>

#include "cpu_info.h"
#include "file_util.h"

#include <QVector>

CpuInfo *CpuInfo::info = nullptr;

CpuInfo* CpuInfo::instance() {

	if ( not info )
		info = new CpuInfo();

	return info;
};

quint8 CpuInfo::getCpuCoreCount() const {

    return sysconf( _SC_NPROCESSORS_ONLN );
};

QList<double> CpuInfo::getLoadAvgs() const {

	return mLoadAvgs;
}

QList<int> CpuInfo::getCpuPercents() const {

	return mCpuPercents;
}

CpuInfo::CpuInfo() : QObject() {

	readTimer.start( 1000, this );

	QStringList lastReadList = FileUtil::readListFromFile( PROC_STAT );
	lastReadList.removeFirst();
	for( int i = 0; i < getCpuCoreCount(); i++ )
		lastRead[ i ] = CpuTimes( lastReadList.value( i ) );
};

void CpuInfo::timerEvent( QTimerEvent *tEvent ) {

	if ( tEvent->timerId() == readTimer.timerId() ) {
		/*
			*
			* Read the load averages
			*
		*/
		QStringList strListAvgs = FileUtil::readStringFromFile( PROC_LOADAVG ).split( QRegExp( "\\s+" ) );
		if ( strListAvgs.count() > 2 ) {
			mLoadAvgs.clear();
			mLoadAvgs << strListAvgs.takeFirst().toDouble();
			mLoadAvgs << strListAvgs.takeFirst().toDouble();
			mLoadAvgs << strListAvgs.takeFirst().toDouble();
		}

		/*
			*
			* Read and compute the CPU usage percentages
			*
		*/
		mCpuPercents.clear();
		QStringList times = FileUtil::readListFromFile( PROC_STAT );

		/* First is the average of all the cores */
		times.removeFirst();

		/* Current values */
		for( int i = 0; i < getCpuCoreCount(); i++ ) {
			thisRead[ i ] = CpuTimes( times.value( i ) );

			qreal totalD = thisRead[ i ].total() - lastRead[ i ].total();
			qreal idleD = thisRead[ i ].idle() - lastRead[ i ].idle();

			mCpuPercents << 100 * ( totalD - idleD ) / totalD;
		}

		lastRead = thisRead;
		thisRead.clear();

		tEvent->accept();
		return;
	}

	QObject::timerEvent( tEvent );
};
