/*
    *
    * This file is a part of Libcsys.
    * Library for managing drive and getting system resource information in real time for CuboCore Application Suite
    * Copyright 2019 CuboCore Group
    *

    *
    * NOTE: Several classes and functions have been used here, which are from LXQt.
    * Below is their copyright header. LXQt functions (modified suitably) use a prefix
    * LXQT_. Any bugs in this entire program are to be reported to us, and not to LXQt.
    *

		* BEGIN_COMMON_COPYRIGHT_HEADER
		* (c)LGPL2+
		*
		* LXQt - a lightweight, Qt based, desktop toolset
		* https://lxqt.org
		*
		* Copyright: 2010-2011 Razor team
		* Authors:
		*   Alexander Sokoloff <sokoloff.a@gmail.com>
		*   Petr Vanek <petr@scribus.info>
		*
		* This program or library is free software; you can redistribute it
		* and/or modify it under the terms of the GNU Lesser General Public
		* License as published by the Free Software Foundation; either
		* version 2.1 of the License, or (at your option) any later version.
		*
		* This library is distributed in the hope that it will be useful,
		* but WITHOUT ANY WARRANTY; without even the implied warranty of
		* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
		* Lesser General Public License for more details.
		*
		* You should have received a copy of the GNU Lesser General
		* Public License along with this library; if not, write to the
		* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
		* Boston, MA 02110-1301 USA
		*
		* END_COMMON_COPYRIGHT_HEADER

	*
	* Other parts of the program are developed independent of LXQt and are
	* released under GLP v3+
	*

    *
    * This program is free software; you can redistribute it and/or modify
    * it under the terms of the GNU General Public License as published by
    * the Free Software Foundation; either version 3 of the License, or
    * (at your option) any later version.
    *

    *
    * This program is distributed in the hope that it will be useful,
    * but WITHOUT ANY WARRANTY; without even the implied warranty of
    * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    * GNU General Public License for more details.
    *

    *
    * You should have received a copy of the GNU General Public License
    * along with this program; if not, vsit http://www.gnu.org/licenses/.
    *
*/


#include "power.h"

#define CONSOLEKIT_SERVICE      "org.freedesktop.ConsoleKit"
#define CONSOLEKIT_PATH         "/org/freedesktop/ConsoleKit/Manager"
#define CONSOLEKIT_INTERFACE    "org.freedesktop.ConsoleKit.Manager"

#define SYSTEMD_SERVICE         "org.freedesktop.login1"
#define SYSTEMD_PATH            "/org/freedesktop/login1"
#define SYSTEMD_INTERFACE       "org.freedesktop.login1.Manager"

static bool LXQT_dbusCallSystemd( QDBusInterface *dbus, const QString &method, bool needBoolArg ) {

    if ( not dbus->isValid() )  {
        qWarning() << "dbusCall: QDBusInterface is invalid" << dbus->service() << dbus->path() << dbus->interface() << method;
        return false;
    }

    QDBusMessage msg = dbus->call( method, needBoolArg ? QVariant( true ) : QVariant() );

    if ( not msg.errorName().isEmpty() ) {
		qWarning() << "Error name " << msg.errorName();
		qWarning() << "Error msg  " << msg.errorMessage();
    }

    // If the method no returns value, we believe that it was successful.
    if ( msg.arguments().isEmpty() or msg.arguments().first().isNull() )
        return true;

    QString response = msg.arguments().first().toString();
    qDebug() << "systemd:" << method << "=" << response;

    return response == QLatin1String("yes") || response == QLatin1String("challenge");
}

Power::Power( QObject *parent ) : QObject( parent ) {

    createPowerProviders();
};

Power::~Power() {

	while ( providers.count() ) {
		providers.takeFirst()->deleteLater();
	}
};

void Power::createPowerProviders() {

    // 1: SystemD

    QDBusInterface *iface = new QDBusInterface( SYSTEMD_SERVICE, SYSTEMD_PATH, SYSTEMD_INTERFACE, QDBusConnection::systemBus() );
    if ( iface->isValid() )
        providers << iface;

    // 2: ConsoleKit

    iface = new QDBusInterface( CONSOLEKIT_SERVICE, CONSOLEKIT_PATH, CONSOLEKIT_INTERFACE, QDBusConnection::systemBus() );
    if ( iface->isValid() )
        providers << iface;
};

bool Power::systemCanSuspend() {

    Q_FOREACH( QDBusInterface *iface, providers ) {
        if ( LXQT_dbusCallSystemd( iface, "CanSuspend", false ) )
			return true;
    }

    return false;
};

bool Power::systemCanHibernate() {

    Q_FOREACH( QDBusInterface *iface, providers ) {
        if ( LXQT_dbusCallSystemd( iface, "CanHibernate", false ) )
			return true;
    }

    return false;
};

bool Power::systemCanReboot() {

    Q_FOREACH( QDBusInterface *iface, providers ) {
        if ( LXQT_dbusCallSystemd( iface, "CanReboot", false ) )
			return true;
    }

    return false;
};

bool Power::systemCanHalt() {

    Q_FOREACH( QDBusInterface *iface, providers ) {
        if ( LXQT_dbusCallSystemd( iface, "CanPowerOff", false ) )
			return true;
    }

    return false;
};

bool Power::systemSuspend() {

    Q_FOREACH( QDBusInterface *iface, providers ) {
        if ( LXQT_dbusCallSystemd( iface, "CanSuspend", false ) )
			return LXQT_dbusCallSystemd( iface, "Suspend", true );
    }

    return false;
};

bool Power::systemHibernate() {

    Q_FOREACH( QDBusInterface *iface, providers ) {
        if ( LXQT_dbusCallSystemd( iface, "CanHibernate", false ) )
			return LXQT_dbusCallSystemd( iface, "Hibernate", true );
    }

    return false;
};

bool Power::systemReboot() {

    Q_FOREACH( QDBusInterface *iface, providers ) {
        if ( LXQT_dbusCallSystemd( iface, "CanReboot", false ) )
			return LXQT_dbusCallSystemd( iface, "Reboot", true );
    }

    return false;
};

bool Power::systemHalt() {

    Q_FOREACH( QDBusInterface *iface, providers ) {
        if ( LXQT_dbusCallSystemd( iface, "CanPowerOff", false ) )
			return LXQT_dbusCallSystemd( iface, "PowerOff", true );
    }

    return false;
};
